import enum

from django.db import models
from django.utils.functional import Promise


__all__ = (
    'JSONField',
    'Choices', 'IntegerChoices', 'TextChoices',
)


if hasattr(models, 'JSONField'):
    JSONField = models.JSONField
else:
    from django.contrib.postgres.fields import JSONField


if hasattr(models, 'IntegerChoices'):
    Choices = models.Choices
    IntegerChoices = models.IntegerChoices
    TextChoices = models.TextChoices
else:
    class ChoicesMeta(enum.EnumMeta):
        """A metaclass for creating a enum choices."""

        def __new__(metacls, classname, bases, classdict, **kwds):
            labels = []
            for key in classdict._member_names:
                value = classdict[key]
                if (
                    isinstance(value, (list, tuple)) and
                    len(value) > 1 and
                    isinstance(value[-1], (Promise, str))
                ):
                    *value, label = value
                    value = tuple(value)
                else:
                    label = key.replace('_', ' ').title()
                labels.append(label)
                # Use dict.__setitem__() to suppress defenses against double
                # assignment in enum's classdict.
                dict.__setitem__(classdict, key, value)
            cls = super().__new__(metacls, classname, bases, classdict, **kwds)
            cls._value2label_map_ = dict(zip(cls._value2member_map_, labels))
            # Add a label property to instances of enum which uses the enum member
            # that is passed in as "self" as the value to use when looking up the
            # label in the choices.
            cls.label = property(lambda self: cls._value2label_map_.get(self.value))
            cls.do_not_call_in_templates = True
            return enum.unique(cls)

        def __contains__(cls, member):
            if not isinstance(member, enum.Enum):
                # Allow non-enums to match against member values.
                return any(x.value == member for x in cls)
            return super().__contains__(member)

        @property
        def names(cls):
            empty = ['__empty__'] if hasattr(cls, '__empty__') else []
            return empty + [member.name for member in cls]

        @property
        def choices(cls):
            empty = [(None, cls.__empty__)] if hasattr(cls, '__empty__') else []
            return empty + [(member.value, member.label) for member in cls]

        @property
        def labels(cls):
            return [label for _, label in cls.choices]

        @property
        def values(cls):
            return [value for value, _ in cls.choices]


    class Choices(enum.Enum, metaclass=ChoicesMeta):
        """Class for creating enumerated choices."""

        def __str__(self):
            """
            Use value when cast to str, so that Choices set as model instance
            attributes are rendered as expected in templates and similar contexts.
            """
            return str(self.value)


    class IntegerChoices(int, Choices):
        """Class for creating enumerated integer choices."""
        pass


    class TextChoices(str, Choices):
        """Class for creating enumerated string choices."""

        def _generate_next_value_(name, start, count, last_values):
            return name



import pytest
from pprint import pprint

from wcd_notifications.services import notifier, manager
from wcd_notifications.models.notifications import Notification
from wcd_notifications.contrib.drf.views import stats_list_view
from wcd_notifications.signals import notifications_sent, notifications_updated

from .utils import mute_signals


@pytest.mark.django_db
def test_stats_collect_mechanics(rf, make_user, django_assert_num_queries):
    user, _ = make_user('1')
    user2, _ = make_user('2')

    with mute_signals(notifications_sent, notifications_updated):
        notifications = notifier.send(
            'Found some {actor} for you {recipient}', [user, user2],
            actor=user,
            flags=[Notification.Readability.UNREAD],
        )
        notifications += notifier.send(
            'Found some {actor} for you {recipient}', [user],
            actor=user2,
            flags=[Notification.Readability.UNREAD],
        )
        notifications += notifier.send(
            'Found some {actor} for you {recipient}', [user],
            actor=user,
            flags=[Notification.Readability.READ],
        )

    with django_assert_num_queries(4):
        stat, = manager.collect_stats([user])

    assert stat.total == 3
    assert stat.flags[Notification.Readability.UNREAD] == 2
    assert stat.flags[Notification.Readability.READ] == 1

    with django_assert_num_queries(4):
        stat2, = manager.collect_stats([user])

    assert stat.pk == stat2.pk


@pytest.mark.django_db
def test_stats_view(rf, make_user, django_assert_num_queries):
    user, _ = make_user('1')
    user2, _ = make_user('2')

    with mute_signals(notifications_sent, notifications_updated):
        notifications = notifier.send(
            'Found some {actor} for you {recipient}', [user, user2],
            actor=user,
            flags=[Notification.Readability.UNREAD],
        )
        notifications += notifier.send(
            'Found some {actor} for you {recipient}', [user],
            actor=user2,
            flags=[Notification.Readability.UNREAD],
        )
    notifications += notifier.send(
        'Found some {actor} for you {recipient}', [user],
        actor=user,
        flags=[Notification.Readability.READ],
    )

    request = rf.get('/stats/')
    request.user = user
    response = stats_list_view(request)

    assert response.status_code == 200
    assert len(response.data) == 1

    stat = response.data[0]

    assert stat['recipient']['id'] == user.pk
    assert stat['total'] == 3
    assert stat['flags'][str(Notification.Readability.UNREAD)] == 2
    assert stat['flags'][str(Notification.Readability.READ)] == 1



import pytest
from pprint import pprint

from wcd_notifications.models import Notification
from wcd_notifications.services import notifier
from wcd_notifications.signals import notifications_sent, notifications_updated

from .utils import mute_signals


@pytest.mark.django_db
def test_simple_send(make_user, django_assert_num_queries):
    user, _ = make_user('1')
    user2, _ = make_user('2')

    with django_assert_num_queries(1):
        with mute_signals(notifications_sent, notifications_updated):
            notifications = notifier.send(
                'Found some {actor} for you {recipient}', [user, user2],
                actor=user,
            )

    assert len(notifications[0].flags) == 1
    assert notifications[0].flags[0] == Notification.Readability.UNREAD

    # Runs with state update:
    with django_assert_num_queries(5):
        notifications = notifier.send(
            'Found some {actor} for you {recipient}', [user, user2],
            actor=user,
        )



import pytest
from pprint import pprint

from django.contrib.auth.models import Permission
from wcd_notifications.services import notifier
from wcd_notifications.contrib.drf.views import notifications_change_flags_view
from wcd_notifications.models.notifications import Notification
from wcd_notifications.utils import get_type_pk_pair, to_content_type_key


@pytest.mark.django_db
def test_list_filtering_view(rf, make_user, django_assert_num_queries):
    user, _ = make_user('1')
    user2, _ = make_user('2')
    user3, _ = make_user('3')
    ct, oid = get_type_pk_pair(user2)
    permission = Permission.objects.create(
        content_type=ct, name='some', codename='some'
    )
    pct, poid = get_type_pk_pair(permission)

    notifications = notifier.send(
        'Found some {actor} for you {recipient}', [user, user2],
        actor=user,
        flags=[Notification.Readability.UNREAD],
    )
    notifications += notifier.send(
        'Found some {actor} for you {recipient}', [user],
        actor=user2,
        flags=[Notification.Readability.UNREAD],
    )

    request = rf.post(
        f'/change-flags/?ids={notifications[2].pk}',
    )
    request.user = user
    response = notifications_change_flags_view(request)

    assert response.status_code == 400

    request = rf.post(
        f'/change-flags/?ids={notifications[2].pk}',
        {'add': [Notification.Readability.READ]},
    )
    request.user = user
    response = notifications_change_flags_view(request)
    mapped = {x: y for x, y in Notification.objects.values_list('pk', 'flags')}

    assert response.status_code == 200
    assert response.data['changed'] == 1
    assert set(mapped[notifications[2].pk]) == {Notification.Readability.READ}

    request = rf.post(
        f'/change-flags/?ids={notifications[2].pk}',
        {'remove': [Notification.Readability.READ]},
    )
    request.user = user
    response = notifications_change_flags_view(request)
    mapped = {x: y for x, y in Notification.objects.values_list('pk', 'flags')}

    assert response.status_code == 200
    assert response.data['changed'] == 1
    assert set(mapped[notifications[2].pk]) == {Notification.Readability.UNREAD}

    request = rf.post(
        f'/change-flags/?ids={notifications[2].pk}',
        {'add': [Notification.Readability.READ]},
    )
    request.user = user
    response = notifications_change_flags_view(request)
    mapped = {x: y for x, y in Notification.objects.values_list('pk', 'flags')}

    assert response.status_code == 200
    assert response.data['changed'] == 1
    assert set(mapped[notifications[2].pk]) == {Notification.Readability.READ}

    request = rf.post('/change-flags/', {'specify': Notification.Readability.READ})
    request.user = user
    response = notifications_change_flags_view(request)
    mapped = {x: y for x, y in Notification.objects.values_list('pk', 'flags')}

    assert response.status_code == 200
    assert response.data['changed'] == 1
    assert set(mapped[notifications[2].pk]) == {Notification.Readability.READ}
    assert set(mapped[notifications[0].pk]) == {Notification.Readability.READ}
    assert set(mapped[notifications[1].pk]) == {Notification.Readability.UNREAD}

    request = rf.post('/change-flags/', {'specify': Notification.Readability.UNREAD})
    request.user = user
    response = notifications_change_flags_view(request)
    mapped = {x: y for x, y in Notification.objects.values_list('pk', 'flags')}

    assert response.status_code == 200
    assert response.data['changed'] == 2



import pytest
from pprint import pprint

from django.contrib.auth.models import Permission
from wcd_notifications.services import notifier
from wcd_notifications.contrib.drf.views import notifications_clear_view
from wcd_notifications.models.notifications import Notification
from wcd_notifications.utils import get_type_pk_pair, to_content_type_key


@pytest.mark.django_db
def test_clear_view(rf, make_user, django_assert_num_queries):
    user, _ = make_user('1')
    user2, _ = make_user('2')
    user3, _ = make_user('3')
    ct, oid = get_type_pk_pair(user2)
    permission = Permission.objects.create(
        content_type=ct, name='some', codename='some'
    )
    pct, poid = get_type_pk_pair(permission)

    notifications = notifier.send(
        'Found some {actor} for you {recipient}', [user, user2],
        actor=user,
        flags=[Notification.Readability.UNREAD],
    )
    notifications += notifier.send(
        'Found some {actor} for you {recipient}', [user],
        actor=user2,
        flags=[Notification.Readability.UNREAD],
    )

    request = rf.post(
        f'/set-flags/?ids={notifications[2].pk}',
        {'flags': Notification.Readability.READ},
    )
    request.user = user
    assert Notification.objects.count() == 3
    response = notifications_clear_view(request)

    assert response.status_code == 200
    assert response.data['cleared'] == 1
    assert Notification.objects.count() == 2



from datetime import datetime
from typing import *
from enum import Enum
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.postgres.fields import ArrayField
from django.contrib.contenttypes.models import ContentType
from django.utils import timezone
from django.utils.translation import pgettext_lazy
from django.db import models
from django.contrib.postgres.indexes import GinIndex, BTreeIndex
from px_pipeline import Filter

from ..utils import (
    Registry, make_generic_Q, to_intarray, ModelDef, content_type_from_id as ctfid
)
from ..compat import IntegerChoices, JSONField


__all__ = (
    'Readability',
    'NotificationQuerySet', 'Notification',
    'make_generic_Q',
    'prepare_ordering',
)
# Phone keyboard:         READ0
READABILITY_BASE_NUMBER = 73230

RecipientDef = Tuple[ContentType, str]


class FlagStats(TypedDict):
    recipient: RecipientDef
    stats: Dict[int, int]


class Readability(IntegerChoices):
    UNREAD = READABILITY_BASE_NUMBER + 2, pgettext_lazy('wcd_notifications', 'Unread')
    READ = READABILITY_BASE_NUMBER + 6, pgettext_lazy('wcd_notifications', 'Read')


class FlagsRegistry(Registry):
    def add(self, choices: Type[Enum]):
        for item in choices:
            assert item.value not in self, (
                f'Value {item.value} already registered as {self[item.value]}, '
                'check your flags for consistency.'
            )
            assert item.value > 9, (
                f'Value {item.value} is reserved for internal use.'
            )

            self[item.value] = item

    register = add


class NotificationQuerySet(models.QuerySet):
    def read(self):
        return self.filter(flags__contains=to_intarray([self.model.Readability.READ]))

    def unread(self):
        return self.filter(flags__contains=to_intarray([self.model.Readability.UNREAD]))

    def recipients(self, recipients: Sequence[ModelDef]):
        return self.filter(make_generic_Q('recipient', recipients))

    def actors(self, actors: Sequence[ModelDef]):
        return self.filter(make_generic_Q('actor', actors))

    def actions(self, actions: Sequence[ModelDef]):
        return self.filter(make_generic_Q('action', actions))

    def targets(self, targets: Sequence[ModelDef]):
        return self.filter(make_generic_Q('target', targets))

    def sent(self, now: Optional[datetime] = None):
        return self.filter(sent_at=now if now is not None else timezone.now())

    def collect_recipients(self):
        fields = 'recipient_content_type', 'recipient_object_id',
        query = (
            self
            .only(*fields)
            .order_by(*fields)
            .distinct(*fields)
            .prefetch_related('recipient')
        )
        return [x.recipient for x in query]

    def collect_total_stats(self) -> Dict[RecipientDef, int]:
        fields = (
            'recipient_content_type', 'recipient_object_id',
            'actor_content_type', 'action_content_type', 'target_content_type',
        )
        records = (
            self
            .values(*fields)
            .order_by()
            .annotate(count=models.Count('pk'))
            .values_list(*fields, 'count')
        )
        return [
            (ctfid(ct), oid, ctfid(actor), ctfid(action), ctfid(target), count)
            for (ct, oid, actor, action, target, count) in records
        ]

    def collect_flag_stats(self) -> List[FlagStats]:
        fields = (
            'recipient_content_type', 'recipient_object_id', 'flag',
            'actor_content_type', 'action_content_type', 'target_content_type',
        )
        records = (
            self
            .annotate(flag=models.Func('flags', function='unnest'))
            .values(*fields)
            .order_by()
            .annotate(count=models.Count('pk'))
            .values_list(*fields, 'count')
        )

        return [
            (ctfid(ct), oid, ctfid(actor), ctfid(action), ctfid(target), fl, count)
            for (ct, oid, fl, actor, action, target, count) in records
        ]


class Notification(models.Model):
    Readability = Readability

    preparator = Filter()
    flags_registry = FlagsRegistry()
    objects: Union[NotificationQuerySet, models.Manager] = NotificationQuerySet.as_manager()

    class Meta:
        verbose_name = pgettext_lazy('wcd_notifications', 'Notification')
        verbose_name_plural = pgettext_lazy('wcd_notifications', 'Notifications')
        ordering = ('-ordering_field', '-sent_at', '-created_at', '-pk',)
        indexes = [
            GinIndex(
                name='wcdnt_%(class)s_flags',
                fields=['flags'], opclasses=('gin__int_ops',)
            ),
            BTreeIndex(fields=['recipient_content_type', 'recipient_object_id']),
            BTreeIndex(fields=['actor_content_type', 'actor_object_id']),
            BTreeIndex(fields=['action_content_type', 'action_object_id']),
            BTreeIndex(fields=['target_content_type', 'target_object_id']),
        ]

    parent = models.ForeignKey(
        'self', null=True, blank=True, on_delete=models.SET_NULL,
        verbose_name=pgettext_lazy('wcd_notifications', 'Parent notification'),
        related_name='children',
    )

    recipient_content_type = models.ForeignKey(
        ContentType,
        verbose_name=pgettext_lazy('wcd_notifications', 'Recipient: Content type'),
        related_name='notification_recipient', on_delete=models.CASCADE,
    )
    recipient_object_id = models.CharField(
        verbose_name=pgettext_lazy('wcd_notifications', 'Recipient: Id'),
        max_length=255, null=False, blank=False,
    )
    recipient = GenericForeignKey('recipient_content_type', 'recipient_object_id')

    actor_content_type = models.ForeignKey(
        ContentType, null=True, blank=True,
        verbose_name=pgettext_lazy('wcd_notifications', 'Actor: Content type'),
        related_name='notification_actor', on_delete=models.SET_NULL,
    )
    actor_object_id = models.CharField(
        verbose_name=pgettext_lazy('wcd_notifications', 'Actor: Id'),
        max_length=255, null=True, blank=True,
    )
    actor = GenericForeignKey('actor_content_type', 'actor_object_id')

    action_content_type = models.ForeignKey(
        ContentType, null=True, blank=True,
        verbose_name=pgettext_lazy('wcd_notifications', 'Action: Content type'),
        related_name='notification_action', on_delete=models.SET_NULL,
    )
    action_object_id = models.CharField(
        verbose_name=pgettext_lazy('wcd_notifications', 'Action: Id'),
        max_length=255, null=True, blank=True,
    )
    action = GenericForeignKey('action_content_type', 'action_object_id')

    target_content_type = models.ForeignKey(
        ContentType, null=True, blank=True,
        verbose_name=pgettext_lazy('wcd_notifications', 'Target: Content type'),
        related_name='notification_target', on_delete=models.SET_NULL,
    )
    target_object_id = models.CharField(
        verbose_name=pgettext_lazy('wcd_notifications', 'Target: Id'),
        max_length=255, null=True, blank=True,
    )
    target = GenericForeignKey('target_content_type', 'target_object_id')

    flags = ArrayField(
        models.PositiveIntegerField(),
        verbose_name=pgettext_lazy('wcd_notifications', 'Flags'),
        null=False, blank=True, default=list,
    )
    message = models.TextField(
        verbose_name=pgettext_lazy('wcd_notifications', 'Message'),
        null=False, blank=False,
    )

    data = JSONField(
        verbose_name=pgettext_lazy('wcd_notifications', 'Data'),
        blank=True, null=False, default=dict,
    )
    sent_at = models.DateTimeField(default=timezone.now, db_index=True)
    created_at = models.DateTimeField(default=timezone.now, db_index=True)
    ordering_field = models.BigIntegerField(db_index=True, editable=False)

    def __str__(self):
        return self.message

    def prepare(
        self,
        force_insert: bool = False,
        force_update: bool = False,
        using: Optional[str] = None,
        update_fields: Optional[Iterable[str]] = None,
    ):
        result = self.preparator({
            'instance': self,
            'force_insert': force_insert,
            'force_update': force_update,
            'using': using,
            'update_fields': update_fields,
        })
        result.setdefault('force_insert', force_insert)
        result.setdefault('force_update', force_update)
        result.setdefault('using', using)
        result.setdefault('update_fields', update_fields)

        return result

    def save(self, **kwargs) -> None:
        prepared = self.prepare(**kwargs)

        return super().save(
            force_insert=prepared['force_insert'],
            force_update=prepared['force_update'],
            using=prepared['using'],
            update_fields=prepared['update_fields'],
        )


Notification.flags_registry.add(Notification.Readability)


def _set_update_field_context(context: dict, value: str):
    if 'update_field' in context and context['update_field']:
        if not isinstance(context['update_field'], list):
            context['update_field'] = list(context['update_field'])

        context['update_field'].append(value)


@Notification.preparator.add
def prepare_ordering(context: dict):
    _set_update_field_context(context, 'ordering_field')
    instance: Notification = context['instance']
    instance.ordering_field = int(instance.sent_at.timestamp() * 1000)

    return context


@Notification.preparator.add
def set_readability_state(context: dict):
    instance: Notification = context['instance']

    if set(instance.flags) & set(x.value for x in Notification.Readability):
        return

    _set_update_field_context(context, 'flags')
    instance.flags.append(Notification.Readability.UNREAD)

    return context



from django.contrib import admin
from django.conf import settings

from .models import Notification, Stats
from .services import manager


formfield_overrides = {}


if 'prettyjson' in settings.INSTALLED_APPS:
    from prettyjson import PrettyJSONWidget
    from .compat import JSONField

    formfield_overrides[JSONField] = {'widget': PrettyJSONWidget}


@admin.register(Notification)
class NotificationAdmin(admin.ModelAdmin):
    list_display = 'id', 'message', 'recipient', 'created_at',
    list_filter = (
        'recipient_content_type', 'actor_content_type',
        'action_content_type', 'target_content_type',
    )
    raw_id_fields = 'parent',
    search_fields = (
        'parent__id',
        'recipient_object_id', 'actor_object_id',
        'action_object_id', 'target_object_id',
        'flags', 'message', 'data',
    )
    formfield_overrides = formfield_overrides

    def save_model(self, request, obj, form, change) -> None:
        result = super().save_model(request, obj, form, change)

        manager.collect_stats([obj.recipient])

        return result


@admin.register(Stats)
class StatsAdmin(admin.ModelAdmin):
    list_display = (
        'id', 'recipient', 'flag', 'amount',
        'recipient_content_type', 'actor_content_type',
        'action_content_type', 'target_content_type',
    )
    list_filter = (
        'recipient_content_type', 'actor_content_type',
        'action_content_type', 'target_content_type',
    )
    list_select_related = list_filter
    search_fields = (
        'recipient_content_type', 'recipient_object_id',
        'flag', 'amount',
    )



from operator import or_
from django import forms
from django.utils.encoding import force_str
from django.contrib.contenttypes.models import ContentType
from django.utils.translation import pgettext_lazy
from django.db import models

from django_filters import rest_framework as filters
from django_filters import fields
from django_filters import widgets
from django_filters.constants import EMPTY_VALUES

from wcd_notifications.models import Notification, Stats
from wcd_notifications.utils import from_content_type_key


class SplitterWidget(widgets.BaseCSVWidget):
    delimiter = ','

    def __init__(self, *args, **kwargs):
        self.delimiter = kwargs.pop('delimiter', self.delimiter)
        super().__init__(*args, **kwargs)

    def value_from_datadict(self, data, files, name):
        try:
            getter = data.getlist
        except AttributeError:
            getter = data.get

        values = getter(name)

        if not isinstance(values, (list, tuple, set)):
            values = [values]

        normalized = (
            [] if x == '' or x is None else x.split(self.delimiter)
            for x in values
        )

        return [
            [y for y in x if y is not None]
            for x in normalized
            if x is not None and len(x) > 0
        ]

    def render(self, name, value, attrs=None, renderer=None):
        if not self._isiterable(value):
            value = [value]

        if len(value) <= 1:
            # delegate to main widget (Select, etc...) if not multiple values
            value = value[0] if value else ""
            return super().render(name, value, attrs, renderer=renderer)

        # if we have multiple values, we need to force render as a text input
        # (otherwise, the additional values are lost)
        value = [force_str(self.surrogate.format_value(v)) for v in value]
        value = self.delimiter.join(list(value))

        return self.surrogate.render(name, value, attrs, renderer=renderer)


class FlatSplitterWidget(SplitterWidget):
    def value_from_datadict(self, data, files, name):
        lists = super().value_from_datadict(data, files, name)

        return [y for x in lists for y in x]


class SplitterField(fields.BaseCSVField):
    widget = SplitterWidget


class FlatSplitterField(SplitterField):
    widget = FlatSplitterWidget


class GenericKeySplitterField(FlatSplitterField):
    type_pk_delimiter = '~'
    default_error_messages = {
        'key_invalid': pgettext_lazy('wcd_notifications', 'Key is invalid.'),
        'no_content_type': pgettext_lazy('wcd_notifications', 'No such content type exists.'),
    }

    def __init__(self, *args, **kwargs):
        self.type_pk_delimiter = kwargs.pop('type_pk_delimiter', self.type_pk_delimiter)
        super().__init__(*args, **kwargs)

    def clean(self, values):
        values = super().clean(values)
        result = []

        for value in values:
            ct, oid, *_ = value.split(self.type_pk_delimiter) + [None, None]

            if not ct or not oid:
                raise forms.ValidationError(
                    self.error_messages['key_invalid'], code='key_invalid',
                )

            try:
                content_type = from_content_type_key(ct)
            except ContentType.DoesNotExist:
                raise forms.ValidationError(
                    self.error_messages['no_content_type'],
                    code='no_content_type',
                )

            result.append((content_type, oid))

        return result


class ContentTypeSplitterField(FlatSplitterField):
    default_error_messages = {
        'no_content_type': pgettext_lazy('wcd_notifications', 'No such content type exists.'),
    }

    def clean(self, values):
        values = super().clean(values)
        result = []

        for value in values:
            if not value or value.isspace():
                if self.required:
                    raise forms.ValidationError(
                        self.error_messages['required'],
                        code='required',
                    )
                continue

            try:
                content_type = from_content_type_key(value)
            except ContentType.DoesNotExist:
                raise forms.ValidationError(
                    self.error_messages['no_content_type'],
                    code='no_content_type',
                )

            result.append(content_type)

        return result


class LookupMultiplicationFilter(filters.BaseInFilter):
    base_field_class = SplitterField
    operator = or_

    def __init__(self, *args, **kwargs):
        self.operator = kwargs.pop('operator', self.operator)
        super().__init__(*args, **kwargs)

    def filter(self, qs, value):
        if value in EMPTY_VALUES:
            return qs
        if self.distinct:
            qs = qs.distinct()
        lookup = "%s__%s" % (self.field_name, self.lookup_expr)

        q = models.Q()

        for element in value:
            q = self.operator(q, models.Q(**{lookup: element}))

        qs = self.get_method(qs)(q)
        return qs


class InFilter(filters.BaseInFilter):
    base_field_class = FlatSplitterField


class NumbersInFilter(InFilter, filters.NumberFilter):
    pass


class QuerySetMethodFilter(InFilter):
    def __init__(self, *args, **kwargs):
        self.qs_method = kwargs.pop('qs_method')
        super().__init__(*args, **kwargs)

    def get_method(self, qs):
        return getattr(qs, self.qs_method)

    def filter(self, qs, value):
        if value in EMPTY_VALUES:
            return qs
        if self.distinct:
            qs = qs.distinct()

        qs = self.get_method(qs)(value)
        return qs


class GenericKeyMethodFilter(QuerySetMethodFilter):
    base_field_class = GenericKeySplitterField


class ContentTypeInFilter(InFilter):
    base_field_class = ContentTypeSplitterField


class BaseFilterSet(filters.FilterSet):
    ids = NumbersInFilter(field_name='pk', lookup_expr='in')


class NotificationsFilterSet(BaseFilterSet):
    actors = GenericKeyMethodFilter(qs_method='actors')
    actions = GenericKeyMethodFilter(qs_method='actions')
    targets = GenericKeyMethodFilter(qs_method='targets')

    actor_types = ContentTypeInFilter(field_name='actor_content_type', lookup_expr='in')
    action_types = ContentTypeInFilter(field_name='action_content_type', lookup_expr='in')
    target_types = ContentTypeInFilter(field_name='target_content_type', lookup_expr='in')

    parents = NumbersInFilter(field_name='parent_id', lookup_expr='in')
    flags = LookupMultiplicationFilter(field_name='flags', lookup_expr='contains')

    sent_at = filters.IsoDateTimeFromToRangeFilter()

    is_sent = filters.BooleanFilter(method='filter_sent')
    is_root = filters.BooleanFilter(method='filter_roots')

    class Meta:
        model = Notification
        fields = [
            'ids',
            'actors', 'actions', 'targets',
            'actor_types', 'action_types', 'target_types',
            'sent_at',
            'is_sent', 'is_root',
        ]

    def filter_sent(self, qs, value):
        if value is not True:
            return qs

        return qs.sent()

    def filter_roots(self, qs, value):
        if value is not True:
            return qs

        return qs.filter(parent_id__isnull=True)


class StatsFilterSet(filters.FilterSet):
    actor_types = ContentTypeInFilter(field_name='actor_content_type', lookup_expr='in')
    action_types = ContentTypeInFilter(field_name='action_content_type', lookup_expr='in')
    target_types = ContentTypeInFilter(field_name='target_content_type', lookup_expr='in')

    class Meta:
        model = Stats
        fields = ['actor_types', 'action_types', 'target_types']



from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('wcd_notifications', '0002_auto_20230616_1153'),
    ]

    operations = [
        migrations.DeleteModel(
            name='Stats',
        ),
    ]



# Generated by Django 2.2.28 on 2023-06-21 11:14

import django.contrib.postgres.indexes
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('wcd_notifications', '0003_delete_stats'),
    ]

    operations = [
        migrations.CreateModel(
            name='Stats',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('recipient_object_id', models.CharField(max_length=255, verbose_name='Recipient: Id')),
                ('flag', models.BigIntegerField(default=0, verbose_name='Flag')),
                ('amount', models.BigIntegerField(default=0, verbose_name='Amount')),
            ],
            options={
                'verbose_name': 'Notification stats',
                'verbose_name_plural': 'List of notifications stats',
                'ordering': ('-pk',),
            },
        ),
        migrations.RemoveIndex(
            model_name='notification',
            name='wcdnt_notification_flags',
        ),
        migrations.AlterField(
            model_name='notification',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AddIndex(
            model_name='notification',
            index=django.contrib.postgres.indexes.GinIndex(fields=['flags'], name='wcdnt_%(class)s_flags', opclasses=('gin__int_ops',)),
        ),
        migrations.AddField(
            model_name='stats',
            name='action_content_type',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='notification_stats_action', to='contenttypes.ContentType', verbose_name='Action: Content type'),
        ),
        migrations.AddField(
            model_name='stats',
            name='actor_content_type',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='notification_stats_actor', to='contenttypes.ContentType', verbose_name='Actor: Content type'),
        ),
        migrations.AddField(
            model_name='stats',
            name='recipient_content_type',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='stats_recipient', to='contenttypes.ContentType', verbose_name='Recipient: Content type'),
        ),
        migrations.AddField(
            model_name='stats',
            name='target_content_type',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='notification_stats_target', to='contenttypes.ContentType', verbose_name='Target: Content type'),
        ),
        migrations.AddIndex(
            model_name='stats',
            index=django.contrib.postgres.indexes.BTreeIndex(fields=['recipient_content_type', 'recipient_object_id'], name='wcd_notific_recipie_7a6270_btree'),
        ),
    ]



from typing import List, Optional, Sequence, Set
from itertools import chain

from django.db import models
from px_pipeline import StraightPipeline

from ..models import Notification, NotificationQuerySet, Stats, TOTAL_FLAG
from ..signals import (
    notifications_flags_changed, notifications_updated, notifications_cleared,
    stats_updated,
)
from ..utils import model_bulk_update_or_create, to_intarray, ModelDef
from ..conf import settings


__all__ = 'change_flags', 'collect_stats', 'clear',


def set_readability_flags_operation(context):
    # FIXME: Nothing works here.
    default_state = Notification.Readability.UNREAD
    states = set(x for x in Notification.Readability)
    added = context['specify'] or context['add']
    to_remove = states - added

    for notification in context['instances']:
        temporary = set(notification.flags) - set(to_remove)

        if len(temporary & states) == 0:
            temporary = [default_state]

        notification.flags = list(temporary)


def _specify_flags(queryset, flags: Set[int]) -> List[Notification]:
    flags = set(flags)
    queryset = queryset.exclude(flags=to_intarray(flags))
    instances: List[Notification] = list(queryset)

    for notification in instances:
        notification.flags = list(flags)

    return instances


def _modify_flags(
    queryset,
    add: Set[int] = set(),
    remove: Set[int] = set()
) -> List[Notification]:
    q = models.Q()

    if len(add) > 0:
        q |= ~models.Q(flags__contains=to_intarray(add))

    if len(remove) > 0:
        q |= models.Q(flags__contains=to_intarray(remove))

    queryset = queryset.filter(q)
    instances: List[Notification] = list(queryset)

    for notification in instances:
        notification.flags = list(
            (set(notification.flags) - remove) | add
        )

    return instances


def change_flags(
    queryset,
    /,
    add: Sequence[int] = [],
    remove: Sequence[int] = [],
    specify: Sequence[int] = [],
) -> List[Notification]:
    assert not (len(add) == 0 and len(remove) == 0 and len(specify) == 0), (
        'Do something. Do not run this with all lists empty.'
    )
    assert not (len(specify) > 0 and (len(add) > 0 or len(remove) > 0)), (
        'Specify parameter overrides any `add` or `remove` actions.'
    )

    instances = []
    add_set = set(add)
    remove_set = set(remove)
    specify_set = set(specify)

    if len(specify) != 0:
        instances = _specify_flags(queryset, flags=specify_set)
    else:
        instances = _modify_flags(queryset, add=add_set, remove=remove_set)

    prepared = StraightPipeline(settings.CHANGE_FLAG_PIPELINE)({
        'instances': instances,
        'add': add_set, 'remove': remove_set, 'specify': specify_set,
        'update_fields': ['flags'],
    })
    instances = prepared['instances']
    update_fields = prepared['update_fields']

    Notification.objects.bulk_update(instances, fields=update_fields)

    notifications_flags_changed.send(
        Notification, instances=instances, add=add_set, remove=remove_set,
        specify=specify_set,
    )
    notifications_updated.send(
        Notification, instances=instances, updated_fields=update_fields,
    )

    return instances


def collect_stats(recipients: Optional[Sequence[ModelDef]]) -> List[Stats]:
    qs: NotificationQuerySet = Notification.objects.all()

    if recipients is not None:
        assert len(recipients) != 0, (
            'Can\'t collect statistics for empty recipients list.'
        )
        qs = qs.recipients(recipients)

    flag_stats = qs.collect_flag_stats()
    total_stats = qs.collect_total_stats()
    stats = chain(flag_stats, (
        (ct, oid, actor, action, target, TOTAL_FLAG, count)
        for (ct, oid, actor, action, target, count) in total_stats
    ))

    instances = model_bulk_update_or_create(Stats, [
        (
            {
                'recipient_content_type': ct,
                'recipient_object_id': oid,
                'actor_content_type': actor,
                'action_content_type': action,
                'target_content_type': target,
                'flag': flag,
            },
            {
                'amount': count,
            }
        )
        for (ct, oid, actor, action, target, flag, count) in stats
    ])
    Stats.objects.recipients(recipients).exclude(pk__in=[x.pk for x in instances]).delete()

    stats_updated.send(Stats, instances=instances)

    return instances


def clear(queryset) -> int:
    instances: List[Notification] = list(queryset)
    # FIXME: Change this to be shure that we're deleting only notifications
    # That received previously.
    deleted, _ = queryset.delete()

    notifications_cleared.send(Notification, instances=instances)

    return deleted



import pytest
from pprint import pprint

from django.contrib.auth.models import Permission
from wcd_notifications.services import notifier
from wcd_notifications.contrib.drf.views import (
    notifications_list_view, flags_list_view
)
from wcd_notifications.models.notifications import Notification
from wcd_notifications.utils import get_type_pk_pair, to_content_type_key


@pytest.mark.django_db
def test_flags_list_view(rf, django_assert_num_queries):
    request = rf.get('/list/')

    with django_assert_num_queries(0):
        response = flags_list_view(request)

    assert response.status_code == 200
    assert len(response.data) == 2
    assert {x['id'] for x in response.data} == {x for x in Notification.Readability}


@pytest.mark.django_db
def test_notifications_list_view(rf, make_user, django_assert_num_queries):
    user, _ = make_user('1')
    request = rf.get('/list/')
    request.user = user
    response = notifications_list_view(request)

    assert response.status_code == 200
    assert len(response.data) == 0

    user2, _ = make_user('2')
    notifications = notifier.send(
        'Found some {actor} for you {recipient}', [user, user2],
        actor=user,
        flags=[Notification.Readability.UNREAD],
    )

    # FIXME: Make them to be only 2.
    with django_assert_num_queries(3):
        response = notifications_list_view(request)

    assert response.status_code == 200
    assert len(response.data) == 1
    assert response.data[0]['message'] == 'Found some 1 for you 1'

    # FIXME: Make them to be only 2.
    request.user = user2
    with django_assert_num_queries(3):
        response = notifications_list_view(request)

    assert response.status_code == 200
    assert len(response.data) == 1
    assert response.data[0]['message'] == 'Found some 1 for you 2'


@pytest.mark.django_db
def test_list_filtering_view(rf, make_user, django_assert_num_queries):
    user, _ = make_user('1')
    user2, _ = make_user('2')
    user3, _ = make_user('3')
    ct, oid = get_type_pk_pair(user2)
    permission = Permission.objects.create(
        content_type=ct, name='some', codename='some'
    )
    pct, poid = get_type_pk_pair(permission)

    notifications = notifier.send(
        'Found some {actor} for you {recipient}', [user, user2],
        actor=user,
        flags=[Notification.Readability.UNREAD],
    )
    notifications += notifier.send(
        'Found some {actor} for you {recipient}', [user],
        actor=user2,
        flags=[Notification.Readability.UNREAD],
    )

    request = rf.get(
        f'/list/'
        f'?ids={notifications[0].pk}&ids={notifications[2].pk}'
    )
    request.user = user
    response = notifications_list_view(request)

    assert response.status_code == 200
    assert len(response.data) == 2

    request = rf.get(f'/list/?ids={notifications[0].pk}')
    request.user = user
    with django_assert_num_queries(3):
        response = notifications_list_view(request)

    assert response.status_code == 200
    assert len(response.data) == 1
    assert response.data[0]['message'] == 'Found some 1 for you 1'

    request = rf.get(
        f'/list/'
        f'?ids={notifications[0].pk}&ids={notifications[2].pk}'
        f'&actors={to_content_type_key(ct)}~{oid}'
    )
    request.user = user
    with django_assert_num_queries(3):
        response = notifications_list_view(request)

    assert response.status_code == 200
    assert len(response.data) == 1
    assert response.data[0]['message'] == 'Found some 2 for you 1'

    notifications += notifier.send(
        'Found some {actor} for you {recipient}', [user],
        actor=permission,
        flags=[Notification.Readability.READ],
    )
    request = rf.get(
        f'/list/'
        f'?actor_types={to_content_type_key(pct)}'
    )
    request.user = user
    with django_assert_num_queries(4):
        response = notifications_list_view(request)

    assert response.status_code == 200
    assert len(response.data) == 1
    assert response.data[0]['message'] == 'Found some auth | user | some for you 1'

    request = rf.get(
        f'/list/'
        f'?sent_at_after={notifications[-2].sent_at.isoformat()}'
    )
    request.user = user
    # FIXME: Must be 2 here:
    with django_assert_num_queries(5):
        response = notifications_list_view(request)

    assert response.status_code == 200
    assert len(response.data) == 2
    assert response.data[0]['message'] == 'Found some auth | user | some for you 1'
    assert response.data[1]['message'] == 'Found some 2 for you 1'

    request = rf.get(
        f'/list/'
        f'?sent_at_before={notifications[-2].sent_at.isoformat()}'
    )
    request.user = user
    # FIXME: Must be 2 here:
    with django_assert_num_queries(3):
        response = notifications_list_view(request)

    assert response.status_code == 200
    assert len(response.data) == 2
    assert response.data[0]['message'] == 'Found some 2 for you 1'
    assert response.data[1]['message'] == 'Found some 1 for you 1'

    request = rf.get(
        f'/list/'
        f'?flags={Notification.Readability.UNREAD}'
    )
    request.user = user
    # FIXME: Must be 2 here:
    with django_assert_num_queries(3):
        response = notifications_list_view(request)

    assert response.status_code == 200
    assert len(response.data) == 2

    # OR operator here:
    request = rf.get(
        f'/list/'
        f'?flags={Notification.Readability.UNREAD}&flags={Notification.Readability.READ}'
    )
    request.user = user
    # FIXME: Must be 2 here:
    with django_assert_num_queries(5):
        response = notifications_list_view(request)

    assert response.status_code == 200
    assert len(response.data) == 3

    # Since there is "and" - it find nothing:
    request = rf.get(
        f'/list/'
        f'?flags={Notification.Readability.UNREAD},{Notification.Readability.READ}'
    )
    request.user = user
    # FIXME: Must be 2 here:
    with django_assert_num_queries(1):
        response = notifications_list_view(request)

    assert response.status_code == 200
    assert len(response.data) == 0



# Generated by Django 2.2.28 on 2023-06-21 12:51

from django.db import migrations

from django.core.management import call_command


def recollect(apps, schema_editor):
    call_command('notifications_collect_stats', verbosity=0)


class Migration(migrations.Migration):

    dependencies = [
        ('wcd_notifications', '0004_auto_20230621_1414'),
    ]

    operations = [
        migrations.RunPython(recollect, migrations.RunPython.noop)
    ]



from dataclasses import dataclass, field
from typing import Sequence
from px_settings.contrib.django import settings as s
from px_pipeline.utils import Executable


__all__ = 'Settings', 'settings',


@s('WCD_NOTIFICATIONS')
@dataclass
class Settings:
    RECIPIENTS_RESOLVER: str = 'wcd_notifications.utils.default_resolve_recipients'
    PREPARATION_PIPELINE: Sequence[Executable] = field(default_factory=list)
    CHANGE_FLAG_PIPELINE: Sequence[Executable] = field(default_factory=lambda: [
        'wcd_notifications.services.manager.set_readability_flags_operation',
    ])


settings = Settings()



from itertools import groupby
from operator import attrgetter
from operator import or_
from django.db import models
from django.contrib.contenttypes.models import ContentType
from django.utils.translation import pgettext_lazy
from functools import lru_cache
from collections import OrderedDict
from typing import Callable, List, Optional, Sequence, Tuple, Type, TypeVar, Union
from django.utils.module_loading import import_string

from .conf import settings


NATURAL_KEY_DELIMITER = '.'
M = TypeVar('M', bound=models.Model)
TypePkPair = Tuple[ContentType, Union[str, int]]
ModelDef = Union[models.Model, TypePkPair]


@lru_cache
def cached_import_string(path: str):
    return import_string(path)


def group_and_sort(items, key=lambda x: x, **kwargs):
    return groupby(sorted(items, key=key, **kwargs), key=key)


def get_type_pk_pair(instance: Optional[models.Model] = None):
    if instance is None:
        return None, None

    return ContentType.objects.get_for_model(instance.__class__), instance.pk


def content_type_from_id(id):
    return ContentType.objects.get_for_id(id)if id is not None else None


def to_content_type_key(ct: ContentType):
    return NATURAL_KEY_DELIMITER.join(ct.natural_key())


def from_content_type_key(key: str) -> ContentType:
    splitted = key.split(NATURAL_KEY_DELIMITER)

    if len(splitted) != 2:
        raise ContentType.DoesNotExist(pgettext_lazy(
            'wcd_notifications', 'No content type found for "{}" key.'
        ).format(key))

    return ContentType.objects.get_by_natural_key(*splitted)


def to_intarray(ints: Sequence[int]) -> str:
    return '{' + ','.join(map(str, ints)) + '}'


class Registry(OrderedDict):
    def add(self, key, value):
        assert key not in self, f'{key} already registered.'

        self[key] = value

    register = add


def resolve_generic_field_data(
    obj: models.Model, field_name: str,
) -> Tuple[Optional[ContentType], Optional[str], Optional[models.Model]]:
    if getattr(obj, f'{field_name}_content_type_id') is None:
        return None, None, None

    return (
        getattr(obj, f'{field_name}_content_type'),
        getattr(obj, f'{field_name}_object_id'),
        getattr(obj, field_name),
    )


def make_generic_Q(
    field_name: str,
    values: Sequence[ModelDef],
    operator: Callable = or_
) -> models.Q:
    q = models.Q()
    ctf = f'{field_name}_content_type'
    oidf = f'{field_name}_object_id'

    for value in values:
        ct, oid = get_type_pk_pair(value) if isinstance(value, models.Model) else value
        q = operator(q, models.Q(**{ctf: ct, oidf: oid}))

    return q


def default_resolve_recipients(request, **kw) -> List[models.Model]:
    user = getattr(request, 'user')

    return [user] if user is not None and user.is_authenticated else []


def resolve_recipients(request, **kw) -> List[models.Model]:
    return cached_import_string(settings.RECIPIENTS_RESOLVER)(request, **kw)


def model_bulk_update_or_create(
    model: Type[M], items: Sequence[Tuple[dict, dict]],
    internal_id_name: str = '_internal_lookup_id',
) -> List[M]:
    if len(items) == 0:
        return []

    id_map = {i + 1: value for i, value in enumerate(items)}
    q = models.Q()
    whens = []
    getid = attrgetter(internal_id_name)

    for id, (filters, _) in id_map.items():
        condition = models.Q(**filters)
        q |= condition
        whens.append(models.When(condition, then=models.Value(id)))

    existing = list(
        model.objects
        .filter(q)
        .annotate(**{internal_id_name: models.Case(
            *whens, default=None, output_field=models.IntegerField(),
        )})
    )
    diff = list(id_map.keys() - {getid(x) for x in existing})

    if len(existing) > 0:
        update_fields = set()

        for instance in existing:
            id = getid(instance)

            for field, value in id_map[id][1].items():
                setattr(instance, field, value)
                update_fields.add(field)

        model.objects.bulk_update(existing, fields=update_fields)

    if len(diff) == 0:
        return existing

    blanks = (
        model(**{**id_map[id][0], **id_map[id][1]})
        for id in diff
    )
    created = model.objects.bulk_create(blanks)

    for i, x in enumerate(created):
        setattr(x, internal_id_name, diff[i])

    return sorted(existing + created, key=getid)



from django.dispatch import Signal


notification_sent = Signal()
notifications_sent = Signal()
notifications_flags_changed = Signal()
notifications_updated = Signal()
notifications_cleared = Signal()

stats_updated = Signal()



from django.apps import AppConfig
from django.utils.translation import pgettext_lazy


__all__ = ('DRFConfig',)


class DRFConfig(AppConfig):
    name = 'wcd_notifications.contrib.drf'
    label = 'wcd_notifications_drf'
    verbose_name = pgettext_lazy('wcd_notifications', 'Notifications DRF')
    default_auto_field = 'django.db.models.BigAutoField'



from django.core.management.base import BaseCommand

from wcd_notifications.services import manager
from wcd_notifications.models import Notification
from wcd_notifications.utils import from_content_type_key


class Command(BaseCommand):
    help = 'Collecting stats in case of some failures.'

    def add_arguments(self, parser):
        parser.add_argument('-r', '--recipients', nargs='+', default=[])

    def handle(self, *args, recipients=[], **options):
        qs = Notification.objects.all()

        if len(recipients) != 0:
            qs = qs.recipients(
                (from_content_type_key(ct), oid) for ct, oid, *_ in (
                    y.split('~') + [None, None] for y in recipients
                )
            )

        resolved_recipients = qs.collect_recipients()

        if len(resolved_recipients) > 0:
            manager.collect_stats(resolved_recipients)

        print(f'Collected statistics for {len(resolved_recipients)} recipients.')



from django.urls import path, include

from rest_framework import status
from rest_framework.response import Response
from rest_framework.generics import ListAPIView, GenericAPIView
from django_filters import rest_framework as filters

from wcd_notifications.models import Notification, Stats
from wcd_notifications.utils import resolve_recipients

from .filters import NotificationsFilterSet, StatsFilterSet
from .serializers import (
    FlagReadSerializer,
    NotificationReadSerializer, NotificationChangeFlagSerializer,
    NotificationClearSerializer,
    StatsReadSerializer,
)


__all__ = (
    'FlagsList',
    'NotificationsList', 'NotificationsChangeFlags',
    'StatsList',
    'flags_list_view',
    'notifications_list_view', 'notifications_change_flags_view',
    'stats_list_view',
    'make_urlpatterns',
)


class ActionApiView(GenericAPIView):
    def post(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        self.perform_action(serializer)
        return Response(serializer.data, status=status.HTTP_200_OK)

    def perform_action(self, serializer):
        serializer.commit()


class RecipientsResolverMixin:
    def get_recipients(self):
        return resolve_recipients(self.request, view=self)

    def get_queryset(self):
        recipients = self.get_recipients()
        qs = super().get_queryset()

        if len(recipients) == 0:
            return qs.none()

        return qs.recipients(recipients)


class NotificationsMixin(RecipientsResolverMixin):
    queryset = (
        Notification.objects
        .select_related(
            'recipient_content_type', 'actor_content_type',
            'action_content_type', 'target_content_type',
        )
        .prefetch_related('recipient', 'actor', 'action', 'target')
    )
    filter_backends = (filters.DjangoFilterBackend,)
    filterset_class = NotificationsFilterSet


class MultiNotificationsActionView(NotificationsMixin, ActionApiView):
    def perform_action(self, serializer):
        queryset = self.filter_queryset(self.get_queryset())
        serializer.commit(queryset)


class FlagsList(ListAPIView):
    serializer_class = FlagReadSerializer
    flags_registry = property(lambda self: Notification.flags_registry)
    # HACK: This is for different api analyzers.
    queryset = Notification.objects.none()

    def list(self, request, *args, **kwargs):
        serializer = self.get_serializer(self.flags_registry.values(), many=True)
        return Response(serializer.data)


class NotificationsList(NotificationsMixin, ListAPIView):
    serializer_class = NotificationReadSerializer


class NotificationsClear(MultiNotificationsActionView):
    serializer_class = NotificationClearSerializer


class NotificationsChangeFlags(MultiNotificationsActionView):
    serializer_class = NotificationChangeFlagSerializer


class StatsList(RecipientsResolverMixin, ListAPIView):
    serializer_class = StatsReadSerializer
    queryset = Stats.objects.all().prefetch_related('recipient')
    filter_backends = (filters.DjangoFilterBackend,)
    filterset_class = StatsFilterSet

    def combine_results(self, queryset):
        return queryset.resolve_total_flags()

    def filter_queryset(self, queryset):
        return self.combine_results(super().filter_queryset(queryset))


flags_list_view = FlagsList.as_view()
notifications_list_view = NotificationsList.as_view()
notifications_change_flags_view = NotificationsChangeFlags.as_view()
notifications_clear_view = NotificationsClear.as_view()
stats_list_view = StatsList.as_view()
app_name = 'wcd_notifications'


def make_urlpatterns(
    flags_list_view=flags_list_view,
    notifications_list_view=notifications_list_view,
    notifications_change_flags_view=notifications_change_flags_view,
    notifications_clear_view=notifications_clear_view,
    stats_list_view=stats_list_view,
):
    return [
        path('flags/', include(([
            path('list/', flags_list_view, name='list'),
        ], app_name), namespace='flags')),
        path('notifications/', include(([
            path('list/', notifications_list_view, name='list'),
            path('change-flags/', notifications_change_flags_view, name='change-flags'),
            path('clear/', notifications_clear_view, name='clear'),
            path('stats/', stats_list_view, name='stats'),
        ], app_name), namespace='notifications')),
    ]



from datetime import datetime
from typing import List, Optional, Sequence
from px_pipeline import StraightPipeline

from django.db import models
from django.utils import timezone

from ..models import Notification
from ..signals import notification_sent, notifications_sent
from ..conf import settings


def send(
    message: str,
    recipients: Sequence[models.Model],
    actor: Optional[models.Model] = None,
    action: Optional[models.Model] = None,
    target: Optional[models.Model] = None,
    parent: Optional[Notification] = None,
    flags: List[int] = [],
    extra: dict = {},
    send_at: Optional[datetime] = None,
    now: Optional[datetime] = None,
    batch_size: Optional[int] = None,
) -> List[Notification]:
    assert len(recipients) > 0, 'Must have at least one recipient.'

    actual_now = timezone.now()
    now = actual_now if now is None else now
    send_at = now if send_at is None else send_at

    notifications = []

    for recipient in recipients:
        notification = Notification(
            message=message.format(
                recipient=recipient, actor=actor, action=action, target=target,
                flags=flags, extra=extra, now=now, actual_now=actual_now,
            ),
            recipient=recipient, actor=actor, action=action, target=target,
            parent=parent, flags=flags, created_at=now, sent_at=send_at,
            data={'actual_now': actual_now.isoformat(), **extra},
        )
        notification.recipient = recipient
        notification.actor = actor
        notification.action = action
        notification.target = target
        notification.prepare()
        notifications.append(notification)

    prepared = StraightPipeline(settings.PREPARATION_PIPELINE)({
        'instances': notifications,
    })
    notifications = Notification.objects.bulk_create(
        prepared['instances'], batch_size=batch_size,
    )

    for notification in notifications:
        notification_sent.send(Notification, instance=notification)

    notifications_sent.send(Notification, instances=notifications)

    return notifications



# Generated by Django 3.2.11 on 2022-10-03 12:47

import django.contrib.postgres.fields
import django.contrib.postgres.indexes
import django.contrib.postgres.operations
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone

from wcd_notifications.compat import JSONField


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        django.contrib.postgres.operations.CreateExtension('intarray'),
        migrations.CreateModel(
            name='Stats',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('recipient_object_id', models.CharField(max_length=255, verbose_name='Recipient: Id')),
                ('flags', JSONField(blank=True, default=dict, verbose_name='Flags stats')),
                ('total', models.BigIntegerField(default=0, verbose_name='Total')),
                ('recipient_content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='stats_recipient', to='contenttypes.contenttype', verbose_name='Recipient: Content type')),
            ],
            options={
                'verbose_name': 'Notification stats',
                'verbose_name_plural': 'List of notifications stats',
                'ordering': ('-pk',),
            },
        ),
        migrations.CreateModel(
            name='Notification',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('recipient_object_id', models.CharField(max_length=255, verbose_name='Recipient: Id')),
                ('actor_object_id', models.CharField(blank=True, max_length=255, null=True, verbose_name='Actor: Id')),
                ('action_object_id', models.CharField(blank=True, max_length=255, null=True, verbose_name='Action: Id')),
                ('target_object_id', models.CharField(blank=True, max_length=255, null=True, verbose_name='Target: Id')),
                ('flags', django.contrib.postgres.fields.ArrayField(base_field=models.PositiveIntegerField(), size=None, verbose_name='Flags')),
                ('message', models.TextField(verbose_name='Message')),
                ('data', JSONField(blank=True, default=dict, verbose_name='Data')),
                ('sended_at', models.DateTimeField(db_index=True, default=django.utils.timezone.now)),
                ('created_at', models.DateTimeField(db_index=True, default=django.utils.timezone.now)),
                ('ordering_field', models.BigIntegerField(db_index=True, editable=False)),
                ('action_content_type', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='notification_action', to='contenttypes.contenttype', verbose_name='Action: Content type')),
                ('actor_content_type', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='notification_actor', to='contenttypes.contenttype', verbose_name='Actor: Content type')),
                ('parent', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='children', to='wcd_notifications.notification', verbose_name='Parent notification')),
                ('recipient_content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='notification_recipient', to='contenttypes.contenttype', verbose_name='Recipient: Content type')),
                ('target_content_type', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='notification_target', to='contenttypes.contenttype', verbose_name='Target: Content type')),
            ],
            options={
                'verbose_name': 'Notification',
                'verbose_name_plural': 'Notifications',
                'ordering': ('ordering_field', '-sended_at', '-created_at', '-pk'),
            },
        ),
        migrations.AddIndex(
            model_name='stats',
            index=django.contrib.postgres.indexes.BTreeIndex(fields=['recipient_content_type', 'recipient_object_id'], name='wcd_notific_recipie_7a6270_btree'),
        ),
        migrations.AddIndex(
            model_name='notification',
            index=django.contrib.postgres.indexes.GinIndex(fields=['flags'], name='wcdnt_notification_flags', opclasses=('gin__int_ops',)),
        ),
        migrations.AddIndex(
            model_name='notification',
            index=django.contrib.postgres.indexes.BTreeIndex(fields=['recipient_content_type', 'recipient_object_id'], name='wcd_notific_recipie_1c0195_btree'),
        ),
        migrations.AddIndex(
            model_name='notification',
            index=django.contrib.postgres.indexes.BTreeIndex(fields=['actor_content_type', 'actor_object_id'], name='wcd_notific_actor_c_ec6f85_btree'),
        ),
        migrations.AddIndex(
            model_name='notification',
            index=django.contrib.postgres.indexes.BTreeIndex(fields=['action_content_type', 'action_object_id'], name='wcd_notific_action__70b4f7_btree'),
        ),
        migrations.AddIndex(
            model_name='notification',
            index=django.contrib.postgres.indexes.BTreeIndex(fields=['target_content_type', 'target_object_id'], name='wcd_notific_target__4dd7da_btree'),
        ),
    ]



from typing import *
from dataclasses import dataclass

from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.utils.translation import pgettext_lazy
from django.db import models
from django.contrib.postgres.indexes import BTreeIndex

from ..utils import (
    make_generic_Q, ModelDef, group_and_sort, resolve_generic_field_data
)


__all__ = 'StatsQuerySet', 'Stats', 'TOTAL_FLAG',

TOTAL_FLAG = 0


# FIXME: This is a fallback model for an old behaviour
@dataclass
class StatsTotal:
    flags: Dict[int, int]

    recipient_content_type_id: Optional[int]
    recipient_content_type: Optional[models.Model]
    recipient_object_id: Optional[str]
    recipient: Optional[models.Model]
    total: Optional[int]


class StatsQuerySet(models.QuerySet):
    def recipients(self, recipients: Sequence[ModelDef]):
        return self.filter(make_generic_Q('recipient', recipients))

    def resolve_total_flags(self):
        items = group_and_sort(self, key=lambda item: (
            item.recipient_content_type_id, item.recipient_object_id
        ))
        result = []

        for _, stats in items:
            # FIXME: This recipient getter is ugly:
            stats = list(stats)
            recipient_content_type, recipient_object_id, recipient = (
                resolve_generic_field_data(stats[0], 'recipient')
            )

            flags = {
                flag: sum(x.amount for x in items)
                for flag, items in group_and_sort(
                    stats, key=lambda item: item.flag
                )
            }
            total = flags.pop(TOTAL_FLAG, None)
            result.append(StatsTotal(
                flags, total=total,
                recipient_content_type_id=getattr(recipient_content_type, 'pk', None),
                recipient_content_type=recipient_content_type,
                recipient_object_id=recipient_object_id,
                recipient=recipient,
            ))

        return result


class Stats(models.Model):
    objects = StatsQuerySet.as_manager()

    class Meta:
        verbose_name = pgettext_lazy('wcd_notifications', 'Notification stats')
        verbose_name_plural = pgettext_lazy(
            'wcd_notifications', 'List of notifications stats'
        )
        ordering = ('-pk',)
        indexes = [
            BTreeIndex(fields=['recipient_content_type', 'recipient_object_id']),
        ]

    recipient_content_type = models.ForeignKey(
        ContentType,
        verbose_name=pgettext_lazy('wcd_notifications', 'Recipient: Content type'),
        related_name='stats_recipient', on_delete=models.CASCADE,
    )
    recipient_object_id = models.CharField(
        verbose_name=pgettext_lazy('wcd_notifications', 'Recipient: Id'),
        max_length=255, null=False, blank=False,
    )
    recipient = GenericForeignKey('recipient_content_type', 'recipient_object_id')

    actor_content_type = models.ForeignKey(
        ContentType, null=True, blank=True,
        verbose_name=pgettext_lazy('wcd_notifications', 'Actor: Content type'),
        related_name='notification_stats_actor', on_delete=models.SET_NULL,
    )
    action_content_type = models.ForeignKey(
        ContentType, null=True, blank=True,
        verbose_name=pgettext_lazy('wcd_notifications', 'Action: Content type'),
        related_name='notification_stats_action', on_delete=models.SET_NULL,
    )
    target_content_type = models.ForeignKey(
        ContentType, null=True, blank=True,
        verbose_name=pgettext_lazy('wcd_notifications', 'Target: Content type'),
        related_name='notification_stats_target', on_delete=models.SET_NULL,
    )

    flag = models.BigIntegerField(
        verbose_name=pgettext_lazy('wcd_notifications', 'Flag'),
        default=0,
    )
    amount = models.BigIntegerField(
        verbose_name=pgettext_lazy('wcd_notifications', 'Amount'),
        default=0,
    )

    def __str__(self):
        return str(self.pk)



# Generated by Django 3.2 on 2023-06-16 11:53

import django.contrib.postgres.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wcd_notifications', '0001_initial'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='notification',
            options={'ordering': ('-ordering_field', '-sent_at', '-created_at', '-pk'), 'verbose_name': 'Notification', 'verbose_name_plural': 'Notifications'},
        ),
        migrations.RenameField(
            model_name='notification',
            old_name='sended_at',
            new_name='sent_at',
        ),
        migrations.AlterField(
            model_name='notification',
            name='flags',
            field=django.contrib.postgres.fields.ArrayField(base_field=models.PositiveIntegerField(), blank=True, default=list, size=None, verbose_name='Flags'),
        ),
    ]



from django.apps import AppConfig
from django.utils.translation import pgettext_lazy


__all__ = ('NotificationsConfig',)


class NotificationsConfig(AppConfig):
    name = 'wcd_notifications'
    verbose_name = pgettext_lazy('wcd_notifications', 'Notifications')
    default_auto_field = 'django.db.models.BigAutoField'

    def ready(self) -> None:
        super().ready()

        from . import subscriptions



from django.utils.translation import pgettext_lazy

from rest_framework import serializers
from rest_framework import exceptions

from wc_shortcodes.registry import CodesRegistry

from wcd_notifications.models import Notification, Stats
from wcd_notifications.utils import (
    Registry, resolve_generic_field_data, to_content_type_key,
)
from wc_shortcodes.transformer import transform
from wcd_notifications.services import manager


__all__ = (
    'FIELD_SERIALIZERS', 'SHORTCODES_REGISTRY',
    'FlagReadSerializer',
    'NotificationReadSerializer', 'NotificationChangeFlagSerializer',
    'NotificationClearSerializer',
    'StatsReadSerializer',
)


FIELD_SERIALIZERS = Registry()
SHORTCODES_REGISTRY = CodesRegistry()


class FlagReadSerializer(serializers.Serializer):
    id = serializers.IntegerField(source='value')
    label = serializers.CharField(source='name')
    display = serializers.SerializerMethodField()

    def get_display(self, obj):
        return getattr(obj, 'label', obj.name)


class DisplayRelatedFieldMixin:
    field_serializers = FIELD_SERIALIZERS
    shortcodes = SHORTCODES_REGISTRY

    def _display_related_field(self, notification: Notification, field_name: str):
        content_type, id, obj = resolve_generic_field_data(notification, field_name)

        if content_type is None or obj is None:
            return None

        result = {
            'id': id,
            'type': to_content_type_key(content_type),
            'display': str(obj),
        }
        model = content_type.model_class()

        if model in self.field_serializers:
            result['props'] = self.field_serializers[model](instance=obj).data

        return result


class NotificationReadSerializer(DisplayRelatedFieldMixin, serializers.ModelSerializer):
    flags_registry = property(lambda self: Notification.flags_registry)

    class Meta:
        model = Notification
        fields = (
            'id',
            'recipient', 'actor', 'action', 'target',
            'flags', 'message', 'data',
            'sent_at', 'created_at',
        )

    recipient = serializers.SerializerMethodField()
    actor = serializers.SerializerMethodField()
    action = serializers.SerializerMethodField()
    target = serializers.SerializerMethodField()

    message = serializers.SerializerMethodField()

    flags = serializers.SerializerMethodField()

    def get_recipient(self, obj):
        return self._display_related_field(obj, 'recipient')

    def get_actor(self, obj):
        return self._display_related_field(obj, 'actor')

    def get_action(self, obj):
        return self._display_related_field(obj, 'action')

    def get_target(self, obj):
        return self._display_related_field(obj, 'target')

    def get_flags(self, obj):
        registry = self.flags_registry

        return [
            (
                FlagReadSerializer(registry[flag]).data
                if flag in registry else
                {'id': flag}
            )
            for flag in obj.flags
        ]

    def get_message(self, obj):
        view = self.context.get('view')

        return transform(self.shortcodes, obj.message, context={
            'instance': obj,
            'view': view,
            'request': getattr(view, 'request', None),
        })


class NotificationChangeFlagSerializer(serializers.Serializer):
    default_error_messages = {
        'all_empty': pgettext_lazy(
            'wcd_notifications',
            'You must pass at least one flag into either `add` or `remove` '
            'or `specify` field.'
        ),
    }

    add = serializers.ListField(
        child=serializers.IntegerField(), allow_empty=True, min_length=0,
        required=False,
    )
    remove = serializers.ListField(
        child=serializers.IntegerField(), allow_empty=True, min_length=0,
        required=False,
    )
    specify = serializers.ListField(
        child=serializers.IntegerField(), allow_empty=True, min_length=0,
        required=False,
    )

    def validate(self, data):
        add = data.get('add') or []
        remove = data.get('remove') or []
        specify = data.get('specify') or []

        if len(add) == 0 and len(remove) == 0 and len(specify) == 0:
            raise exceptions.ValidationError(
                self.error_messages['all_empty'], 'all_empty'
            )

        return data

    def commit(self, queryset):
        notifications = manager.change_flags(
            queryset,
            add=self.validated_data.get('add') or [],
            remove=self.validated_data.get('remove') or [],
            specify=self.validated_data.get('specify') or [],
        )

        self._data = {'changed': len(notifications)}


class NotificationClearSerializer(serializers.Serializer):
    def commit(self, queryset):
        cleared = manager.clear(queryset)

        self._data = {'cleared': cleared}


class StatsReadSerializer(DisplayRelatedFieldMixin, serializers.Serializer):
    recipient = serializers.SerializerMethodField()
    flags = serializers.JSONField()
    total = serializers.IntegerField()

    def get_recipient(self, obj):
        return self._display_related_field(obj, 'recipient')



from django.dispatch import receiver

from .signals import (
    notifications_sent, notifications_updated, notifications_cleared,
)
from .services import manager


@receiver(notifications_sent)
@receiver(notifications_updated)
@receiver(notifications_cleared)
def update_stats_on_notifications_change(sender, instances, **kw):
    recipients = {x.recipient for x in instances}

    if len(recipients) > 0:
        manager.collect_stats(recipients)
